@@ -1,6 +1,7 @@
 import argparse
 import base64
 import csv
+import io
 import json
 import logging
 import math
@@ -13,7 +14,7 @@ import traceback
 from getpass import getpass
 from http.server import SimpleHTTPRequestHandler, ThreadingHTTPServer
 from io import BytesIO
-from multiprocessing.dummy import Pool
+from multiprocessing import Pool
 from pathlib import Path
 from time import sleep
 
@@ -25,6 +26,7 @@ import requests
 from pydicom import uid
 from pydicom.encaps import encapsulate, generate_pixel_data_frame
 from pydicom.errors import InvalidDicomError
+from pydicom.filewriter import write_file_meta_info
 from pydicom.pixel_data_handlers import apply_color_lut
 from pynetdicom import AE, evt, ALL_TRANSFER_SYNTAXES
 from pynetdicom import sop_class
@@ -40,16 +42,43 @@ logger = logging.getLogger('echolog')
 OS_VAR_PREFIX = 'US2_'
 
 
+def setup_logging(filename=None, verbose=None):
+    if verbose:
+        logging.basicConfig()
+        logging.getLogger().setLevel(logging.DEBUG)
+    if filename:
+        fh = logging.FileHandler(filename)
+        logging.getLogger().addHandler(fh)
+
+
 class ENV:
     def __init__(self, env):
-        prefix = "" if env and env == "production" else env + "-"
+        prefix = f'{env}-'.replace('production-', '')
         self.cloud = '://' not in env
         self.api_url = f"https://{prefix}api.us2.ai" if self.cloud else env
 
 
-class Auth:
-    auth_mutex = threading.Lock()
+def call_repeatedly_fail_safe(interval, func, *args, timeout=5):
+    stopped = threading.Event()
 
+    def loop():
+        while not stopped.wait(interval):  # the first call is in `interval` secs
+            logger.error('Hello there')
+            while True:
+                try:
+                    func(*args)
+                    break
+                except Exception as exc:
+                    logger.error(f'Failed to call {func.__name__} due to {exc}, retrying in {timeout}s')
+                    sleep(timeout)
+
+    t = threading.Thread(target=loop)
+    t.daemon = True
+    t.start()
+    return stopped.set
+
+
+class Auth:
     def __init__(self, args, username, password):
         self.args = args
         self.api_url = args.env.api_url
@@ -62,9 +91,11 @@ class Auth:
         self.groups = []
         self.sub = ''
         self.authenticate()
+        time_till_exp = self.exp - time.time()
+        call_repeatedly_fail_safe(time_till_exp * 3 / 4, self.refresh)
 
     def expired(self):
-        return time.time() > self.exp - 5
+        return time.time() > self.exp
 
     def unpack_auth(self, request):
         data = unpack(request)
@@ -82,16 +113,14 @@ class Auth:
             requests.post(f"{self.api_url}/users/login", json={'username': self.username, 'password': self.password}),
         )
 
-    def check_token(self):
-        if self.expired():
-            with self.auth_mutex:
-                try:
-                    self.unpack_auth(
-                        requests.post(f"{self.api_url}/users/refresh", json={'refreshToken': self.refresh_token}),
-                    )
-                except Exception as exc:
-                    logger.warning(f'Failed to renew token due to {exc}, re-authenticating')
-                    self.authenticate()
+    def refresh(self):
+        try:
+            self.unpack_auth(
+                requests.post(f"{self.api_url}/users/refresh", json={'refreshToken': self.refresh_token}),
+            )
+        except Exception as exc:
+            logger.warning(f'Failed to renew token due to {exc}, re-authenticating')
+            self.authenticate()
 
     def customer(self):
         groups = self.groups
@@ -102,7 +131,6 @@ class Auth:
             return self.sub
 
     def get_headers(self):
-        self.check_token()
         return {"Authorization": f"Bearer {self.id_token}"}
 
 
@@ -197,6 +225,8 @@ def pad_to_multiple(arr, size, dims=(1, 2)):
 def package_dicom(ds, anonymize, compress):
     # Populate required values for file meta information
     ds.remove_private_tags()
+    if not anonymize and not compress:
+        return
     media = parse_dicom_pixel(ds)
     if anonymize:
         ds.walk(person_data_callback)
@@ -277,28 +307,72 @@ def create_proxy(args):
     return Proxy
 
 
+def initializer(filename, verbose):
+    setup_logging(filename, verbose)
+
+
 class Handler(FileSystemEventHandler):
     def __init__(self, args):
         self.args = args
-        self.pool = Pool(args.n)
+        self.pool = Pool(args.n, initializer=initializer, initargs=(args.log, args.verbose))
         self.pbar = tqdm(total=0)
-        self.closed = False
-        self.write(['customer', 'trial', 'patientID', 'visit', 'filename'])
+        if self.args.csv_out:
+            csv.writer(open(self.args.csv_out, 'a', newline='')).writerow(
+                ['customer', 'trial', 'patientID', 'visit', 'filename'])
+
+    def __getstate__(self):
+        self_dict = self.__dict__.copy()
+        del self_dict['pool']
+        del self_dict['pbar']
+        return self_dict
+
+    def __setstate__(self, state):
+        self.__dict__.update(state)
 
     def processing(self):
         return self.pbar.n < self.pbar.total
 
-    def write(self, row):
-        if self.args.csv_out:
-            csv.writer(open(self.args.csv_out, 'a', newline='')).writerow(row)
-
     def stop(self):
-        self.closed = True
         self.pool.terminate()
 
     def join(self):
         self.pool.join()
 
+    def handle_err(self, err, vargs, kwargs):
+        if isinstance(err, InvalidDicomError):
+            logger.debug(f'Error during process call: {err}, vargs: {vargs}, kwargs: {kwargs}')
+            self.pbar.total -= 1
+            self.pbar.update(0)
+            return
+        logger.error(f'Error during process call: {err}, vargs: {vargs}, kwargs: {kwargs}')
+        logger.debug(traceback.format_tb(err.__traceback__))
+        self.pbar.update(1)
+
+    def handle(self, *vargs, **kwargs):
+        self.pbar.total += 1
+        self.pbar.refresh()
+        self.pbar.disable = False
+        worker = HandlerWorker(self, *vargs, **kwargs)
+        self.pool.apply_async(worker.process,
+                              error_callback=lambda err: self.handle_err(err, vargs, kwargs),
+                              callback=lambda _: self.pbar.update(1))
+
+    def on_any_event(self, event):
+        logger.debug(f'got event {event}')
+        if isinstance(event, (FileCreatedEvent, FileMovedEvent, FileModifiedEvent)):
+            path = event.src_path
+            wait_file(path)
+            self.handle(path)
+
+
+class HandlerWorker:
+    def __init__(self, handler: Handler, path=None, ds=None, ae_title=None, f=None):
+        self.path = path
+        self.ds = ds
+        self.ae_title = ae_title
+        self.f = f
+        self.args = handler.args
+
     def file_params(self, ds, ae_title, filename=None):
         customer = getattr(self.args, 'customer', '') or ae_title
         stem = '_'.join(map(str, filter(bool, [ds.SOPInstanceUID, filename])))
@@ -311,6 +385,7 @@ class Handler(FileSystemEventHandler):
         }
 
     def upload(self, ds, param):
+        logger.info(f'uploading {param}')
         content_type = 'application/dicom'
         auth = self.args.auth
         env = self.args.env
@@ -330,67 +405,49 @@ class Handler(FileSystemEventHandler):
         buf.seek(0)
         return unpack(requests.put(url, data=buf.read(), headers=headers, params=upload_param))
 
-    def process(self, path=None, ds=None, ae_title=None):
-        self.pbar.disable = False
-        if self.closed:
-            return
-        path = path and Path(path)
-        params = self.file_params(ds or pydicom.dcmread(path, stop_before_pixels=True), ae_title=ae_title,
-                                  filename=getattr(path, 'stem', None))
-        extracted = self.args.extracted
-        k = tuple(params.values())
-        if k not in extracted:
-            ds = ds or pydicom.dcmread(path)
-            if self.args.filter and not any(
-                    getattr(ds, k, None) == v for f in self.args.filter for k, v in [f.split('=', 1)]):
-                logger.info(f'Skipping {ds.SOPInstanceUID} due to no matching filter')
-                return
-            package_dicom(ds, anonymize=self.args.anonymize, compress=self.args.compress)
-            dst = self.args.dst
-            if dst:
-                src = self.args.src
-                rel = path.relative_to(src) if src else f"{ds.SOPInstanceUID}.dcm"
-                out = (Path(dst) / rel).with_suffix(".dcm")
-                if self.args.overwrite or not out.is_file():
-                    out.parent.mkdir(exist_ok=True, parents=True)
-                    ds.save_as(out)
-            if hasattr(self.args, "auth"):
-                self.upload(ds, params)
-            extracted.add(k)
-            self.write(k)
-        self.pbar.update(1)
-
-    def handle_err(self, err, vargs, kwargs):
-        if isinstance(err, InvalidDicomError):
-            logger.debug(f'Error during process call: {err}, vargs: {vargs}, kwargs: {kwargs}')
-            self.pbar.total -= 1
-            self.pbar.update(0)
-            return
-        logger.error(f'Error during process call: {err}, vargs: {vargs}, kwargs: {kwargs}')
-        logger.debug(traceback.format_tb(err.__traceback__))
-        self.pbar.update(1)
-
-    def handle(self, *vargs, **kwargs):
-        self.pbar.total += 1
-        self.pbar.refresh()
-        self.pool.apply_async(self.process, vargs, kwargs,
-                              error_callback=lambda err: self.handle_err(err, vargs, kwargs))
-
-    def on_any_event(self, event):
-        logger.debug(f'got event {event}')
-        if isinstance(event, (FileCreatedEvent, FileMovedEvent, FileModifiedEvent)):
-            path = event.src_path
-            wait_file(path)
-            self.handle(path)
+    def process(self):
+        try:
+            path = self.path and Path(self.path)
+            ds = self.ds or pydicom.dcmread(self.f or path)
+            params = self.file_params(ds, ae_title=self.ae_title, filename=getattr(path, 'stem', None))
+            logger.info(f'processing {params}')
+            extracted = self.args.extracted
+            k = tuple(params.values())
+            if k not in extracted:
+                if self.args.filter and not any(
+                        getattr(ds, k, None) == v for f in self.args.filter for k, v in [f.split('=', 1)]):
+                    logger.info(f'Skipping {ds.SOPInstanceUID} due to no matching filter')
+                    return
+                package_dicom(ds, anonymize=self.args.anonymize, compress=self.args.compress)
+                dst = self.args.dst
+                if dst:
+                    src = self.args.src
+                    rel = path.relative_to(src) if src else f"{ds.SOPInstanceUID}.dcm"
+                    out = (Path(dst) / rel).with_suffix(".dcm")
+                    if self.args.overwrite or not out.is_file():
+                        out.parent.mkdir(exist_ok=True, parents=True)
+                        ds.save_as(out)
+                if hasattr(self.args, "auth"):
+                    self.upload(ds, params)
+                extracted.add(k)
+                if self.args.csv_out:
+                    csv.writer(open(self.args.csv_out, 'a', newline='')).writerow(k)
+        finally:
+            if self.f:
+                self.f.close()
 
 
 def handle_store(event, handler):
     """Handle EVT_C_STORE events."""
-    remote_ae_title = event.assoc.remote['ae_title'].strip().decode()
-    ds = event.dataset
-    ds.file_meta = event.file_meta
-    ds.preamble = b"\0" * 128
-    handler.handle(ds=ds, ae_title=remote_ae_title)
+    f = io.BytesIO()
+    f.write(b'\x00' * 128)
+    f.write(b'DICM')
+    # Encode and write the File Meta Information
+    write_file_meta_info(f, event.file_meta)
+    # Write the encoded dataset
+    f.write(event.request.DataSet.getvalue())
+    f.seek(0)
+    handler.handle(f=f, ae_title=event.assoc.remote['ae_title'].strip().decode())
     return 0x0000
 
 
@@ -457,7 +514,7 @@ def read_args():
         help="Path to csv file of extracted cases")
     parser.add_argument(
         "--verbose", action='store_true',
-        help="Path to csv file of extracted cases")
+        help="Enable all application debug/info logs")
     parser.add_argument(
         '--no-anonymization', action='store_false', dest='anonymize',
         help="No anonymization of data prior to upload",
@@ -544,6 +601,11 @@ def read_args():
         default=[],
         nargs='+',
     )
+    parser.add_argument(
+        '--src-glob',
+        help='glob expression to apply on src directory',
+        default='*',
+    )
     args = parser.parse_args(sys.argv[1:])
     if args.config_file.is_file():
         with args.config_file.open() as f:
@@ -553,12 +615,7 @@ def read_args():
     args.src = args.src and Path(args.src)
     args.extracted = (args.extracted and extracted_list(args.extracted)) or set()
     args.sync = [PacsConnection(c) for c in args.sync]
-    if args.verbose:
-        logging.basicConfig()
-        logging.getLogger().setLevel(logging.DEBUG)
-    if args.log:
-        fh = logging.FileHandler(args.log)
-        logging.getLogger().addHandler(fh)
+    setup_logging(args.log, args.verbose)
     if args.upload or args.sync or args.proxy:
         args.auth = Auth(
             args,
@@ -576,7 +633,7 @@ def main():
     try:
         if args.src:
             src = args.src
-            paths = [src] if src.is_file() else [p for p in src.rglob("*") if p.is_file()]
+            paths = [src] if src.is_file() else [p for p in src.rglob(args.src_glob) if p.is_file()]
             args.n = len(paths)
             args.i = 0
             for path in paths:
@@ -597,6 +654,7 @@ def main():
             ae = AE(args.pacs_ae_title)
             for sop in args.pacs_sops:
                 ae.add_supported_context(sop, ALL_TRANSFER_SYNTAXES)
+            ae.maximum_pdu_size = 0
             pacs = ae.start_server(('0.0.0.0', args.pacs_port), block=False, evt_handlers=handlers)
         if args.proxy:
             httpd = ThreadingHTTPServer(server_address=('0.0.0.0', args.proxy), RequestHandlerClass=create_proxy(args))
